@@ -11,20 +11,22 @@ import genshi.template
 from lxml import etree
 from sql import Literal
 
+from trytond.config import config
+from trytond.i18n import gettext
 from trytond.pool import PoolMeta, Pool
 from trytond.model import (ModelSQL, ModelView, Workflow, fields, dualmethod,
     Unique)
+from trytond.model.exceptions import AccessError
 from trytond.pyson import Eval, If
 from trytond.transaction import Transaction
-from trytond.tools import reduce_ids, grouped_slice
+from trytond.tools import reduce_ids, grouped_slice, sortable_values
 from trytond import backend
+
+from trytond.modules.account_payment.exceptions import ProcessError
 from trytond.modules.company import CompanyReport
 
 from .sepa_handler import CAMT054
 
-__all__ = ['Journal', 'Group', 'Payment', 'Mandate', 'Message',
-    'MandateReport']
-
 # XXX fix: https://genshi.edgewall.org/ticket/582
 from genshi.template.astutil import ASTCodeGenerator, ASTTransformer
 if not hasattr(ASTCodeGenerator, 'visit_NameConstant'):
@@ -42,11 +44,31 @@ if not hasattr(ASTTransformer, 'visit_NameConstant'):
     # Re-use visit_Name because _clone is deleted
     ASTTransformer.visit_NameConstant = ASTTransformer.visit_Name
 
+if config.getboolean('account_payment_sepa', 'filestore', default=False):
+    file_id = 'message_file_id'
+    store_prefix = config.get(
+        'account_payment_sepa', 'store_prefix', default=None)
+else:
+    file_id = None
+    store_prefix = None
+
+INITIATOR_IDS = [
+    (None, ''),
+    ('eu_at_02', "SEPA Creditor Identifier"),
+    ('be_vat', "Belgian Enterprise Number"),
+    ('es_nif', "Spanish VAT Number"),
+    ]
+
 
 class Journal(metaclass=PoolMeta):
     __name__ = 'account.payment.journal'
-    company_party = fields.Function(fields.Many2One('party.party',
-            'Company Party'), 'on_change_with_company_party')
+    company_party = fields.Function(fields.Many2One(
+            'party.party', "Company Party",
+            context={
+                'company': Eval('company', -1),
+                },
+            depends=['company']),
+        'on_change_with_company_party')
     sepa_bank_account_number = fields.Many2One('bank.account.number',
         'Bank Account Number', states={
             'required': Eval('process_method') == 'sepa',
@@ -79,6 +101,20 @@ class Journal(metaclass=PoolMeta):
             },
         translate=False,
         depends=['process_method'])
+    sepa_payable_initiator_id = fields.Selection(
+        INITIATOR_IDS, "SEPA Payable Initiator Identifier",
+        states={
+            'invisible': Eval('process_method') != 'sepa',
+            },
+        depends=['process_method'],
+        help="The identifier used for the initiating party.")
+    sepa_receivable_initiator_id = fields.Selection(
+        INITIATOR_IDS, "SEPA Receivable Initiator Identifier",
+        states={
+            'invisible': Eval('process_method') != 'sepa',
+            },
+        depends=['process_method'],
+        help="The identifier used for the initiating party.")
     sepa_batch_booking = fields.Boolean('Batch Booking', states={
             'invisible': Eval('process_method') != 'sepa',
             },
@@ -101,6 +137,20 @@ class Journal(metaclass=PoolMeta):
         if sepa_method not in cls.process_method.selection:
             cls.process_method.selection.append(sepa_method)
 
+    @classmethod
+    def __register__(cls, module_name):
+        cursor = Transaction().connection.cursor()
+        sql_table = cls.__table__()
+        super().__register__(module_name)
+
+        # Migration from 5.4: sepa identifier merged into eu_at_02
+        for name in {'payable', 'receivable'}:
+            column = getattr(sql_table, 'sepa_%s_initiator_id' % name)
+            cursor.execute(*sql_table.update(
+                    columns=[column],
+                    values=['eu_at_02'],
+                    where=column == 'sepa'))
+
     @classmethod
     def default_company_party(cls):
         pool = Pool()
@@ -144,10 +194,6 @@ class Group(metaclass=PoolMeta):
     @classmethod
     def __setup__(cls):
         super(Group, cls).__setup__()
-        cls._error_messages.update({
-                'no_mandate': 'No valid mandate for payment "%s"',
-                'no_bank_account_number': 'No bank account for payment "%s".',
-                })
         cls._buttons.update({
                 'generate_message': {},
                 })
@@ -165,7 +211,10 @@ class Group(metaclass=PoolMeta):
             mandates = Payment.get_sepa_mandates(self.payments)
             for payment, mandate in zip(self.payments, mandates):
                 if not mandate:
-                    self.raise_user_error('no_mandate', payment.rec_name)
+                    raise ProcessError(
+                        gettext('account_payment_sepa'
+                            '.msg_payment_process_no_mandate',
+                            payment=payment.rec_name))
                 # Write one by one because mandate.sequence_type must be
                 # recomputed each time
                 Payment.write([payment], {
@@ -175,8 +224,10 @@ class Group(metaclass=PoolMeta):
         else:
             for payment in self.payments:
                 if not payment.sepa_bank_account_number:
-                    self.raise_user_error(
-                        'no_bank_account_number', payment.rec_name)
+                    raise ProcessError(
+                        gettext('account_payment_sepa'
+                            '.msg_payment_process_no_iban',
+                            payment=payment.rec_name))
         self.generate_message(_save=False)
 
     @dualmethod
@@ -192,7 +243,7 @@ class Group(metaclass=PoolMeta):
                 group.sepa_messages = ()
             message = tmpl.generate(group=group,
                 datetime=datetime, normalize=unicodedata.normalize,
-                ).filter(remove_comment).render()
+                ).filter(remove_comment).render().encode('utf8')
             message = Message(message=message, type='out', state='waiting',
                 company=group.company)
             group.sepa_messages += (message,)
@@ -222,36 +273,52 @@ class Group(metaclass=PoolMeta):
         Payment = pool.get('account.payment')
         keyfunc = self.sepa_group_payment_key
         # re-browse to align cache
-        payments = Payment.browse(sorted(self.payments, key=keyfunc))
+        payments = Payment.browse(sorted(
+                self.payments, key=sortable_values(keyfunc)))
         for key, grouped_payments in groupby(payments, key=keyfunc):
             yield dict(key), list(grouped_payments)
 
+    @property
+    def sepa_message_id(self):
+        return self.number
+
 
 class Payment(metaclass=PoolMeta):
     __name__ = 'account.payment'
 
     sepa_mandate = fields.Many2One('account.payment.sepa.mandate', 'Mandate',
         ondelete='RESTRICT',
+        states={
+            'readonly': Eval('state') != 'draft',
+            'invisible': ((Eval('process_method') != 'sepa')
+                | (Eval('kind') != 'receivable')),
+            },
         domain=[
             ('party', '=', Eval('party', -1)),
             ('company', '=', Eval('company', -1)),
+            If(Eval('state') == 'draft',
+                ('state', '=', 'validated'),
+                (),
+                )
             ],
-        depends=['party', 'company'])
+        depends=['party', 'company', 'state', 'process_method', 'kind'])
     sepa_mandate_sequence_type = fields.Char('Mandate Sequence Type',
         readonly=True)
     sepa_return_reason_code = fields.Char('Return Reason Code', readonly=True,
         states={
-            'invisible': (~Eval('sepa_return_reason_code')
-                & (Eval('state') != 'failed')),
+            'invisible': ((Eval('process_method') != 'sepa')
+                | (~Eval('sepa_return_reason_code')
+                    & (Eval('state') != 'failed'))),
             },
-        depends=['state'])
+        depends=['process_method', 'state'])
     sepa_return_reason_information = fields.Text('Return Reason Information',
         readonly=True,
         states={
-            'invisible': (~Eval('sepa_return_reason_information')
-                & (Eval('state') != 'failed')),
+            'invisible': ((Eval('process_method') != 'sepa')
+                | (~Eval('sepa_return_reason_information')
+                    & (Eval('state') != 'failed'))),
             },
-        depends=['state'])
+        depends=['process_method', 'state'])
     sepa_end_to_end_id = fields.Function(fields.Char('SEPA End To End ID'),
         'get_sepa_end_to_end_id', searcher='search_end_to_end_id')
     sepa_instruction_id = fields.Function(fields.Char('SEPA Instruction ID'),
@@ -304,8 +371,8 @@ class Payment(metaclass=PoolMeta):
     def sepa_remittance_information(self):
         if self.description:
             return self.description
-        elif self.line and self.line.origin:
-            return getattr(self.line.origin, 'rec_name', '')
+        elif self.line and self.line.move_origin:
+            return getattr(self.line.move_origin, 'rec_name', '')
 
     @property
     def sepa_bank_account_number(self):
@@ -329,6 +396,15 @@ class Payment(metaclass=PoolMeta):
             date = Transaction().context.get('date_value')
         return super(Payment, self).create_clearing_move(date=date)
 
+    @classmethod
+    def view_attributes(cls):
+        return super().view_attributes() + [
+            ('//separator[@id="sepa_return_reason"]', 'states', {
+                    'invisible': ((Eval('process_method') != 'sepa')
+                        | (Eval('state') != 'failed')),
+                    }),
+            ]
+
 
 class Mandate(Workflow, ModelSQL, ModelView):
     'SEPA Mandate'
@@ -336,13 +412,18 @@ class Mandate(Workflow, ModelSQL, ModelView):
     party = fields.Many2One('party.party', 'Party', required=True, select=True,
         states={
             'readonly': Eval('state').in_(
-                ['requested', 'validated', 'canceled']),
+                ['requested', 'validated', 'cancelled']),
             },
-        depends=['state'])
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['state', 'company'])
     account_number = fields.Many2One('bank.account.number', 'Account Number',
         ondelete='RESTRICT',
         states={
-            'readonly': Eval('state').in_(['validated', 'canceled']),
+            'readonly': (
+                Eval('state').in_(['validated', 'cancelled'])
+                | ~Eval('party')),
             'required': Eval('state') == 'validated',
             },
         domain=[
@@ -358,12 +439,8 @@ class Mandate(Workflow, ModelSQL, ModelView):
         depends=['state', 'identification_readonly'])
     identification_readonly = fields.Function(fields.Boolean(
             'Identification Readonly'), 'get_identification_readonly')
-    company = fields.Many2One('company.company', 'Company', required=True,
-        select=True,
-        domain=[
-            ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', -1)),
-            ],
+    company = fields.Many2One(
+        'company.company', "Company", required=True, select=True,
         states={
             'readonly': Eval('state') != 'draft',
             },
@@ -373,7 +450,7 @@ class Mandate(Workflow, ModelSQL, ModelView):
             ('one-off', 'One-off'),
             ], 'Type',
         states={
-            'readonly': Eval('state').in_(['validated', 'canceled']),
+            'readonly': Eval('state').in_(['validated', 'cancelled']),
             },
         depends=['state'])
     sequence_type_rcur = fields.Boolean(
@@ -387,13 +464,13 @@ class Mandate(Workflow, ModelSQL, ModelView):
             ('B2B', 'Business to Business'),
             ], 'Scheme', required=True,
         states={
-            'readonly': Eval('state').in_(['validated', 'canceled']),
+            'readonly': Eval('state').in_(['validated', 'cancelled']),
             },
         depends=['state'])
     scheme_string = scheme.translated('scheme')
     signature_date = fields.Date('Signature Date',
         states={
-            'readonly': Eval('state').in_(['validated', 'canceled']),
+            'readonly': Eval('state').in_(['validated', 'cancelled']),
             'required': Eval('state') == 'validated',
             },
         depends=['state'])
@@ -401,11 +478,11 @@ class Mandate(Workflow, ModelSQL, ModelView):
             ('draft', 'Draft'),
             ('requested', 'Requested'),
             ('validated', 'Validated'),
-            ('canceled', 'Canceled'),
+            ('cancelled', 'Cancelled'),
             ], 'State', readonly=True)
     payments = fields.One2Many('account.payment', 'sepa_mandate', 'Payments')
     has_payments = fields.Function(fields.Boolean('Has Payments'),
-        'has_payments')
+        'get_has_payments')
 
     @classmethod
     def __setup__(cls):
@@ -413,8 +490,8 @@ class Mandate(Workflow, ModelSQL, ModelView):
         cls._transitions |= set((
                 ('draft', 'requested'),
                 ('requested', 'validated'),
-                ('validated', 'canceled'),
-                ('requested', 'canceled'),
+                ('validated', 'cancelled'),
+                ('requested', 'cancelled'),
                 ('requested', 'draft'),
                 ))
         cls._buttons.update({
@@ -439,18 +516,30 @@ class Mandate(Workflow, ModelSQL, ModelView):
         t = cls.__table__()
         cls._sql_constraints = [
             ('identification_unique', Unique(t, t.company, t.identification),
-                'The identification of the SEPA mandate must be unique '
-                'in a company.'),
+                'account_payment_sepa.msg_mandate_unique_id'),
             ]
-        cls._error_messages.update({
-                'delete_draft_canceled': ('You can not delete mandate "%s" '
-                    'because it is not in draft or canceled state.'),
-                })
+
+    @classmethod
+    def __register__(cls, module_name):
+        cursor = Transaction().connection.cursor()
+        table = cls.__table__()
+
+        super().__register__(module_name)
+
+        # Migration from 5.6: rename state canceled to cancelled
+        cursor.execute(*table.update(
+                [table.state], ['cancelled'],
+                where=table.state == 'canceled'))
 
     @staticmethod
     def default_company():
         return Transaction().context.get('company')
 
+    @fields.depends('company')
+    def on_change_company(self):
+        self.identification_readonly = self.default_identification_readonly(
+            company=self.company.id if self.company else None)
+
     @staticmethod
     def default_type():
         return 'recurrent'
@@ -467,38 +556,49 @@ class Mandate(Workflow, ModelSQL, ModelView):
     def default_state():
         return 'draft'
 
-    @staticmethod
-    def default_identification_readonly():
+    @classmethod
+    def default_identification_readonly(cls, **pattern):
         pool = Pool()
         Configuration = pool.get('account.configuration')
         config = Configuration(1)
-        return bool(config.sepa_mandate_sequence)
+        return bool(config.get_multivalue('sepa_mandate_sequence', **pattern))
 
     def get_identification_readonly(self, name):
         return bool(self.identification)
 
     def get_rec_name(self, name):
+        name = '(%s)' % self.id
         if self.identification:
-            return self.identification
-        return '(%s)' % self.id
+            name = self.identification
+        if self.account_number:
+            name += ' @ %s' % self.account_number.rec_name
+        return name
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        return [tuple(('identification',)) + tuple(clause[1:])]
+        if clause[1].startswith('!') or clause[1].startswith('not '):
+            bool_op = 'AND'
+        else:
+            bool_op = 'OR'
+        return [bool_op,
+            ('identification',) + tuple(clause[1:]),
+            ('account_number',) + tuple(clause[1:]),
+            ]
 
     @classmethod
     def create(cls, vlist):
         pool = Pool()
-        Sequence = pool.get('ir.sequence')
         Configuration = pool.get('account.configuration')
 
         config = Configuration(1)
         vlist = [v.copy() for v in vlist]
+        default_company = cls.default_company()
         for values in vlist:
             if (config.sepa_mandate_sequence
                     and not values.get('identification')):
-                values['identification'] = Sequence.get_id(
-                    config.sepa_mandate_sequence.id)
+                values['identification'] = config.get_multivalue(
+                    'sepa_mandate_sequence',
+                    company=values.get('company', default_company)).get()
             # Prevent raising false unique constraint
             if values.get('identification') == '':
                 values['identification'] = None
@@ -550,10 +650,10 @@ class Mandate(Workflow, ModelSQL, ModelView):
             return 'RCUR'
 
     @classmethod
-    def has_payments(cls, mandates, name):
+    def get_has_payments(cls, mandates, name):
         pool = Pool()
         Payment = pool.get('account.payment')
-        payment = Payment.__table__
+        payment = Payment.__table__()
         cursor = Transaction().connection.cursor()
 
         has_payments = dict.fromkeys([m.id for m in mandates], False)
@@ -562,9 +662,9 @@ class Mandate(Workflow, ModelSQL, ModelView):
             cursor.execute(*payment.select(payment.sepa_mandate, Literal(True),
                     where=red_sql,
                     group_by=payment.sepa_mandate))
-            has_payments.update(cursor.fetchall())
+            has_payments.update(cursor)
 
-        return {'has_payments': has_payments}
+        return has_payments
 
     @classmethod
     @ModelView.button
@@ -586,16 +686,19 @@ class Mandate(Workflow, ModelSQL, ModelView):
 
     @classmethod
     @ModelView.button
-    @Workflow.transition('canceled')
+    @Workflow.transition('cancelled')
     def cancel(cls, mandates):
-        # TODO must be automaticaly canceled 13 months after last collection
+        # TODO must be automaticaly cancelled 13 months after last collection
         pass
 
     @classmethod
     def delete(cls, mandates):
         for mandate in mandates:
-            if mandate.state not in ('draft', 'canceled'):
-                cls.raise_user_error('delete_draft_canceled', mandate.rec_name)
+            if mandate.state not in ('draft', 'cancelled'):
+                raise AccessError(
+                    gettext('account_payment_sepa'
+                        '.msg_mandate_delete_draft_cancelled',
+                        mandate=mandate.rec_name))
         super(Mandate, cls).delete(mandates)
 
 
@@ -610,18 +713,17 @@ class Message(Workflow, ModelSQL, ModelView):
         'readonly': Eval('state') != 'draft',
         }
     _depends = ['state']
-    message = fields.Text('Message', states=_states, depends=_depends)
+    message = fields.Binary('Message', filename='filename',
+        file_id=file_id, store_prefix=store_prefix,
+        states=_states, depends=_depends)
+    message_file_id = fields.Char("Message File ID", readonly=True)
     filename = fields.Function(fields.Char('Filename'), 'get_filename')
     type = fields.Selection([
             ('in', 'IN'),
             ('out', 'OUT'),
             ], 'Type', required=True, states=_states, depends=_depends)
-    company = fields.Many2One('company.company', 'Company', required=True,
-        select=True,
-        domain=[
-            ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', -1)),
-            ],
+    company = fields.Many2One(
+        'company.company', "Company", required=True, select=True,
         states={
             'readonly': Eval('state') != 'draft',
             },
@@ -632,7 +734,7 @@ class Message(Workflow, ModelSQL, ModelView):
             ('draft', 'Draft'),
             ('waiting', 'Waiting'),
             ('done', 'Done'),
-            ('canceled', 'Canceled'),
+            ('cancelled', 'Cancelled'),
             ], 'State', readonly=True, select=True)
 
     @classmethod
@@ -642,8 +744,8 @@ class Message(Workflow, ModelSQL, ModelView):
             ('draft', 'waiting'),
             ('waiting', 'done'),
             ('waiting', 'draft'),
-            ('draft', 'canceled'),
-            ('waiting', 'canceled'),
+            ('draft', 'cancelled'),
+            ('waiting', 'cancelled'),
             }
         cls._buttons.update({
                 'cancel': {
@@ -666,30 +768,37 @@ class Message(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        cursor = Transaction().connection.cursor()
         pool = Pool()
         Group = pool.get('account.payment.group')
+        transaction = Transaction()
+        cursor = transaction.connection.cursor()
+        update = transaction.connection.cursor()
+        table = cls.__table__()
 
         super(Message, cls).__register__(module_name)
 
         # Migration from 3.2
-        if TableHandler.table_exist(Group._table):
+        if backend.TableHandler.table_exist(Group._table):
             group_table = Group.__table_handler__(module_name)
             if group_table.column_exist('sepa_message'):
                 group = Group.__table__()
-                table = cls.__table__()
                 cursor.execute(*group.select(
                         group.id, group.sepa_message, group.company))
-                for group_id, message, company_id in cursor.fetchall():
-                    cursor.execute(*table.insert(
+                for group_id, message, company_id in cursor:
+                    update.execute(*table.insert(
                             [table.message, table.type, table.company,
                                 table.origin, table.state],
-                            [[message, 'out', company_id,
+                            [[
+                                    message, 'out', company_id,
                                     'account.payment.group,%s' % group_id,
                                     'done']]))
                 group_table.drop_column('sepa_message')
 
+        # Migration from 5.6: rename state canceled to cancelled
+        cursor.execute(*table.update(
+                [table.state], ['cancelled'],
+                where=table.state == 'canceled'))
+
     @staticmethod
     def default_type():
         return 'in'
@@ -716,11 +825,9 @@ class Message(Workflow, ModelSQL, ModelView):
     @classmethod
     def get_origin(cls):
         IrModel = Pool().get('ir.model')
+        get_name = IrModel.get_name
         models = cls._get_origin()
-        models = IrModel.search([
-                ('model', 'in', models),
-                ])
-        return [(None, '')] + [(m.model, m.name) for m in models]
+        return [(None, '')] + [(m, get_name(m)) for m in models]
 
     @classmethod
     @ModelView.button
@@ -746,7 +853,7 @@ class Message(Workflow, ModelSQL, ModelView):
 
     @classmethod
     @ModelView.button
-    @Workflow.transition('canceled')
+    @Workflow.transition('cancelled')
     def cancel(cls, messages):
         pass
 
@@ -774,9 +881,8 @@ class Message(Workflow, ModelSQL, ModelView):
                 return tag.namespace
 
     def parse(self):
-        message = self.message.encode('utf-8')
-        f = BytesIO(message)
-        namespace = self.get_namespace(message)
+        f = BytesIO(self.message)
+        namespace = self.get_namespace(self.message)
         handlers = self._get_handlers()
         if namespace not in handlers:
             raise  # TODO UserError
